/***********************************************************************

  FinalMix: DeviceOSS.cpp
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  0.9.2		Initial
  0.9.3		Low latency, two worker threads
  0.9.4		Complete rewrite, native pthreads instead of wxThread

***********************************************************************/

#include "generic.h"
#include "DeviceOSS.h"

#if defined(__WXGTK__) || defined(__WXMOTIF__)

#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

void * play_thread(void * data) {
	OSSDevice * parent = (OSSDevice * ) data;
	samples_t second = 2 * 2 * 44100; // FIXME
	
	while (parent->opened) {
		if (!parent->playing)
			sem_wait(&parent->play_sem);
			
		pthread_mutex_lock(&parent->buf_mutex);
			
		size_t tw;
		size_t wr;
		size_t buf_max = parent->latency;
		size_t len = buf_max;
		size_t rlen = len;
		size_t played;
		bool push;

		while ((rlen > 0) && (parent->buf_size > 0)) {
			tw = min(rlen, min(parent->buf_len - parent->buf_start, parent->buf_size));
			wr = write(parent->pdev, (char * ) parent->buf + parent->buf_start, tw);
			parent->buf_start += wr;
			parent->buf_size -= wr;
			rlen -= wr;
			if (parent->buf_start == parent->buf_len)
				parent->buf_start = 0;
		}
		
		played = len - rlen;
		parent->position += played;
		push = (parent->buf_size < (parent->buf_len / 2));
			
		pthread_mutex_unlock(&parent->buf_mutex);
		
		if (push)
			sem_post(&parent->push_sem);

		struct timespec req;

		if (buf_max >= second) {
			req.tv_sec = 1;
			req.tv_nsec = 0;
			nanosleep(&req, NULL);
		} else {
			normsample_t flt = (normsample_t) 1000000000 * buf_max / second / 2;
			if (flt > 0) {
				req.tv_sec = 0;
				req.tv_nsec = (long) flt;
				nanosleep(&req, NULL);
			}
		}
	}
	
	return NULL;
}

void * push_thread(void * data) {
	OSSDevice * parent = (OSSDevice * ) data;
	
	while (true) {
		sem_wait(&parent->push_sem);
		
		size_t tr;
		size_t len;
		
		pthread_mutex_lock(&parent->buf_mutex);
		
		tr = parent->buf_len - parent->buf_size;
		
		if (tr > 0) {
			len = parent->cb->OnBuffer(parent->tmpbuf, tr / SampleLen(parent->pfmt) / parent->pfmt.channels, (parent->position + parent->buf_size) / SampleLen(parent->pfmt) / parent->pfmt.channels, parent->pfmt) * SampleLen(parent->pfmt) * parent->pfmt.channels;
			if (len > 0) {
		
				size_t tr;
				size_t rd;
				size_t off = 0;
				size_t tlen = len;
			
				if ((parent->buf_size < parent->buf_len) && (parent->buf_start + parent->buf_size < parent->buf_len)) {
					tr = parent->buf_len - parent->buf_start - parent->buf_size;
					rd = min(tr, tlen);
					memcpy((char * ) parent->buf + parent->buf_start + parent->buf_size, (char * ) parent->tmpbuf + off, rd);
					parent->buf_size += rd;
					tlen -= rd;
					off += rd;
				}
				if ((parent->buf_size < parent->buf_len) && (parent->buf_start + parent->buf_size >= parent->buf_len) && (parent->buf_start > 0)) {
					tr = parent->buf_len - parent->buf_size;
					rd = min(tr, tlen);
					memcpy((char * ) parent->buf + parent->buf_start + parent->buf_size - parent->buf_len, (char * ) parent->tmpbuf + off, rd);
					parent->buf_size += rd;
				}
				
			}
		}
		
		pthread_mutex_unlock(&parent->buf_mutex);
	}
	
	return NULL;
}

OSSDevice::OSSDevice() {
	pthread_mutexattr_t mattr;

	buf = NULL;
	tmpbuf = NULL;
	pdev = -1;
	opened = false;
	
	pthread_mutexattr_init(&mattr);
	pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_FAST_NP);
	pthread_mutex_init(&buf_mutex, &mattr);
	
	pthread_mutexattr_destroy(&mattr);
}

OSSDevice::~OSSDevice() {
	done();
	
	pthread_mutex_destroy(&buf_mutex);
}

wxString OSSDevice::Name() const {
	return("*nix OSS device");
}

bool OSSDevice::Init(const format_ex_t fmt, DeviceCallback & callback) {
	pthread_attr_t tattr;

	done();
	
	cb = &callback;
	pfmt = fmt;
	
	buf_count = 10; // FIXME configure
	latency = 10240; // FIXME configure, must be a multiply of 1024
	buf_len = latency * buf_count;
	
	buf_start = 0;
	buf_size = 0;
	position = 0;
	opened = false;
	playing = false;
	
	pdev = open("/dev/dsp", O_WRONLY); // FIXME config
	if (pdev == -1) {
		done();
		return false;
	}
	
/*	int param = (latency / 1024) << 16 + 10;
	param = 0x000A000A; // FIXME use addaptive value from above
	if (ioctl(pdev, SNDCTL_DSP_SETFRAGMENT, &param) == -1) {
		done();
		return false;
	}*/
		
	// FIXME check for correct setting
	
	int format = AFMT_S16_LE; // FIXME real param
	if (ioctl(pdev, SNDCTL_DSP_SETFMT, &format) == -1) {
		done();
		return false;
	}

	int stereo = 1; // FIXME real param
	if (ioctl(pdev, SNDCTL_DSP_STEREO, &stereo) == -1) {
		done();
		return false;
	}
		
	int rate = 44100; // FIXME real param
	if (ioctl(pdev, SNDCTL_DSP_SPEED, &rate) == -1) {
		done();
		return false;
	}
	
	buf = new char[buf_len];
	if (!buf) {
		done();
		return false;
	}
	
	tmpbuf = new char[buf_len];
	if (!tmpbuf) {
		done();
		return false;
	}
	
	opened = true;

	sem_init(&play_sem, 0, 0);	
	sem_init(&push_sem, 0, 0);	

	pthread_attr_init(&tattr);
	pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);
	pthread_create(&play, &tattr, play_thread, this);
	pthread_create(&push, &tattr, push_thread, this);
	pthread_attr_destroy(&tattr);
	
	return true;
}

bool OSSDevice::Play() {
	playing = true;
	sem_post(&push_sem);
	sem_post(&play_sem);
	return true;
}

bool OSSDevice::Pause() {
	if (playing)
		playing = false;
	else {
		playing = true;
		sem_post(&push_sem);
		sem_post(&play_sem);
	}
	return true;
}

bool OSSDevice::Stop() {
	if (playing)
		playing = false;
		
	done();
	
	return true;
}

bool OSSDevice::Done() {
	if (playing)
		playing = false;

	done();
	
	return true;
}

samples_t OSSDevice::GetPosition() {
	if (opened) {											/* is device initialized? */
		pthread_mutex_lock(&buf_mutex);						/* enter buffer critical section */
		samples_t pos = position;
		pthread_mutex_unlock(&buf_mutex);					/* leave buffer critical section */
		return (pos / SampleLen(pfmt));						/* position in samples */
	} else
		return 0;
}

Device * OSSDevice::RegisterDevice() {
	return (new OSSDevice);
}

void OSSDevice::done() {
	pthread_mutex_lock(&buf_mutex);
	
	// FIXME check if created
	pthread_cancel(play);
	pthread_cancel(push);
	
	sem_destroy(&push_sem);
	sem_destroy(&play_sem);
	
	if (tmpbuf) {
		delete[] (char * ) tmpbuf;
		buf = NULL;
	}
	
	if (buf) {
		delete[] (char * ) buf;
		buf = NULL;
	}
	
	opened = false;
	if (pdev != -1)
		close(pdev);
		
	pthread_mutex_unlock(&buf_mutex);
}

#endif
