/***********************************************************************

  FinalMix: Frame.cpp
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  0.9.0		Initial version
  0.9.1		Cleanup
  0.9.2		Beginning with peak file

***********************************************************************/

#include "generic.h"
#include "Frame.h"
#include "Draw.h"
#include "Dialogs.h"
#include "Project.h"
#include "Peaks.h"
#include <wx/filedlg.h>
#include <wx/busyinfo.h>
#include <wx/progdlg.h>
#include <wx/filename.h>
#include <wx/tipdlg.h>

#include "xpm/finalmix.xpm"

#include "xpm/playbut.xpm"
#include "xpm/pausebut.xpm"
#include "xpm/stopbut.xpm"

#define CONTROLS 150
#define TRACK_LEFT 10
#define TRACK_RIGHT 10
#define TRACK_RAW 2
#define TRACK_HEIGHT 256
#define TRACK_SPACE 10

Frame::Frame(const wxString & title)
	: wxFrame((wxFrame *) NULL, -1, title) {
	
	SetIcon(wxIcon(finalmix_xpm));    
	configure = wxConfigBase::Get();
	
	SetSize(configure->Read("PositionX", 10), configure->Read("PositionY", 10), configure->Read("Width", 600), configure->Read("Height", 400));
	
	menuFile = new wxMenu;
	menuEdit = new wxMenu;
	submenuPaste = new wxMenu;
	menuTracks = new wxMenu;
	menuOptions = new wxMenu;
	menuHelp = new wxMenu;
	
	menuFile->Append(ID_New, "&New...", "Create new project");
	menuFile->Append(ID_Open, "&Open...\tCtrl+O", "Open project from file");
	menuFile->Append(ID_Save, "&Save\tCtrl+S", "Save project");
	menuFile->Append(ID_SaveAs, "S&ave as...", "Save project to a different file");
	menuFile->Append(ID_Revert, "&Revert", "Revert to the state of previous save");
	menuFile->AppendSeparator();
	menuFile->Append(ID_Render, "R&ender...", "Create final output");
	menuFile->AppendSeparator();
	menuFile->Append(ID_Quit, "&Quit\tCtrl+Q", "Quit " FM_NAME);
	
	submenuPaste->Append(ID_PasteInto, "&Paste into", "Paste from clipboard at cursor");
	submenuPaste->Append(ID_PasteOver, "P&aste over", "Paste from clipboard over cursor");
	submenuPaste->Append(ID_PasteNewTrack, "Paste as &new track", "Paste from clipboard as new track");
	
	menuEdit->Append(ID_Undo, "&Undo\tCtrl+U", "Undo last action");
	menuEdit->Append(ID_Redo, "&Redo", "Repeat undone action");
	menuEdit->AppendSeparator();
	menuEdit->Append(ID_Cut, "&Cut", "Cut selection to clipboard");
	menuEdit->Append(ID_Copy, "C&opy\tCtrl+C", "Copy selection to clipboard");
	menuEdit->Append(ID_PasteMenu, "&Paste", submenuPaste);
	
	menuTracks->Append(ID_Add, "&Add", "Add new track");
	menuTracks->Append(ID_Import, "&Import...", "Import track from audio file");
	menuTracks->Append(ID_Flatten, "&Flatten", "Make all changes permanent");
	
	menuOptions->Append(ID_Settings, "&Settings...", "Configure " FM_NAME);
	
	menuHelp->Append(ID_Help, "&Help...\tF1", "See help");
	menuHelp->Append(ID_Tips, "&Tips...", "Tips of the day");
	menuHelp->AppendSeparator();
	menuHelp->Append(ID_About, "&About...", "About " FM_NAME);
	
	menuBar = new wxMenuBar(wxMB_DOCKABLE);
	menuBar->Append(menuFile, "&File");
	menuBar->Append(menuEdit, "&Edit");
	menuBar->Append(menuTracks, "&Tracks");
	menuBar->Append(menuOptions, "&Options");
	menuBar->Append(menuHelp, "&Help");
	
	SetMenuBar(menuBar);
	
	CreateStatusBar();
	SetStatusText(FM_NAME " " FM_VERSION_STRING);
	
	toolBar = CreateToolBar(wxTB_DOCKABLE | wxTB_HORIZONTAL | wxTB_3DBUTTONS, -1, "ToolBar");
	playBut = new wxBitmapButton(toolBar, ID_PlayBut, wxBitmap(playbut_xpm));
	toolBar->AddControl(playBut);
	pauseBut = new wxBitmapButton(toolBar, ID_PauseBut, wxBitmap(pausebut_xpm));
	toolBar->AddControl(pauseBut);
	stopBut = new wxBitmapButton(toolBar, ID_StopBut, wxBitmap(stopbut_xpm));
	toolBar->AddControl(stopBut);
	toolBar->Realize();

	controls = new Controls(this);
	area = new Area(this, controls);
	timer = new wxTimer(this, ID_Timer);
	project.SetCursorVisual(false);
	timer->Start(40);
	
	cbox = new wxBoxSizer(wxHORIZONTAL);
	cbox->Add(controls, 0, wxGROW, 0, NULL);
	cbox->Add(area, 1, wxGROW, 0, NULL);
	
	SetSizer(cbox);
		
	project.SetPlayDevice(project.GetDevice(0)); // choose device from saved configuration
}

Frame::~Frame(void) {
	wxRect fr = GetRect();
	
	configure->Write("PositionX", fr.x);
	configure->Write("PositionY", fr.y);
	configure->Write("Width", fr.width);
	configure->Write("Height", fr.height);
	configure->Flush();
}

BEGIN_EVENT_TABLE(Frame, wxFrame)
	EVT_MENU(ID_Quit, Frame::OnQuit)
	EVT_MENU(ID_About, Frame::OnAbout)
	EVT_MENU(ID_Tips, Frame::OnTips)
	EVT_MENU(ID_Open, Frame::OnOpen)
	EVT_MENU(ID_SaveAs, Frame::OnSaveAs)
	EVT_MENU(ID_Import, Frame::OnImport)
	
	EVT_BUTTON(ID_PlayBut, Frame::OnPlayBut)
	EVT_BUTTON(ID_PauseBut, Frame::OnPauseBut)
	EVT_BUTTON(ID_StopBut, Frame::OnStopBut)
	
	EVT_TIMER(ID_Timer, Frame::OnTimer)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(Area, wxScrolledWindow)
	EVT_PAINT(Area::OnPaint)
	EVT_SCROLLWIN(Area::OnScroll)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(Controls, wxWindow)
END_EVENT_TABLE()

void Frame::OnQuit(wxCommandEvent & WXUNUSED(event)) {
	if (project.Changed()) {
	}
	Close(true);
}
    
void Frame::OnAbout(wxCommandEvent & WXUNUSED(event)) {
	// about dialog
}

class TipProvider : public wxTipProvider {
	public:
		TipProvider(size_t currentTip) 
			: wxTipProvider(currentTip) {};
		
		wxString GetTip() { 
			wxString tip = "This is a beta version of " FM_NAME ". Please be kind and ignore all the bugs and missing features, they will be fixed and implemented soon.";
			return(tip);
		};
		
		size_t GetCurrentTip() const {
			return(0);
		};
};

void Frame::ShowTips(bool force) {
	if ((configure->Read("ShowTips", true)) || (force)) {
		size_t index = configure->Read("CurrentTip", 0L);
		TipProvider * provider = new TipProvider(index);
		
		bool showtips = wxShowTip(this,provider);
		index = provider->GetCurrentTip();
		
		configure->Write("ShowTips", showtips);
		configure->Write("CurrentTip", (int) index);
		
		delete provider;
	}
}

void Frame::OnTips(wxCommandEvent & WXUNUSED(event)) {
	ShowTips(true);
}
    
void Frame::OnOpen(wxCommandEvent & WXUNUSED(event)) {
	wxString filename = wxFileSelector("Open project", configure->Read("/dlgdirs/Open",""), "", "fmf", FM_NAME " files (*.fmf)|*.fmf|all files|*", wxOPEN | wxHIDE_READONLY | wxFILE_MUST_EXIST, this);
	
	if (!filename.empty()) {
		wxString path;
		wxFileName::SplitPath(filename, &path, NULL, NULL);
		configure->Write("/dlgdirs/Open", path);
	}
}
    
void Frame::OnSaveAs(wxCommandEvent & WXUNUSED(event)) {
	wxString filename = wxFileSelector("Save project as", configure->Read("/dlgdirs/SaveAs",""), "", "fmf", FM_NAME " files (*.fmf)|*.fmf|all files|*", wxSAVE | wxHIDE_READONLY | wxOVERWRITE_PROMPT, this);
	
	if (!filename.empty()) {
		wxString path;
		wxFileName::SplitPath(filename, &path, NULL, NULL);
		configure->Write("/dlgdirs/SaveAs", path);
	}
}
    
void Frame::OnImport(wxCommandEvent & WXUNUSED(event)) {
	wxString filename = wxFileSelector("Import track(s)", configure->Read("/dlgdirs/Import",""), "", "wav", "Wave files (*.wav)|*.wav|AIFF files (*.aif?)|*.aif?|all files|*", wxOPEN | wxHIDE_READONLY | wxFILE_MUST_EXIST, this);
	
	if (!filename.empty()) {
		wxFileName fname(filename);
		ASSERT((fname.IsOk()) && (fname.FileExists()));
		wxDateTime modtime = fname.GetModificationTime();
		wxString path;
		wxFileName::SplitPath(filename, &path, NULL, NULL);
		configure->Write("/dlgdirs/Import", path);
		
		wxWindowDisabler disableAll;
		
		wxFile * file = new wxFile(filename, wxFile::read_write); 
		Loader * ld;
		index_t i;
		bool closed = false;
		
		if (file->IsOpened()) {
			enum status_t ret = UNK_FORMAT;
			for (i = 0, ld = project.GetLoader(i); ld != NULL; i++, ld = project.GetLoader(i)) {
				ret = ld->Attach(file, filename, modtime);
				if (ret == OK)
					break;
				else if (ret == IO_ERROR)
					wxLogWarning(ld->Name() + ": Input/output error");
				else if (ret == UNK_ERROR)
					wxLogWarning(ld->Name() + ": Internal loader error");
				else if (ret == STRUCT_ERR)
					wxLogWarning(ld->Name() + ": Error in file structure");
				else if (ret == UNS_FORMAT)
					wxLogMessage(ld->Name() + ": Unsupported file format");
				else if (ret == UNK_FORMAT)
					wxLogMessage(ld->Name() + ": Unrecognized file format");
				else
					wxLogError(ld->Name() + ": Unknown return value (assertion failed)");
			}
			
	 		switch (ret) {
				case OK:
					closed = true;
					if (ld->NumIntTracks() == 0)
						wxMessageBox(filename + " contains no audio tracks/channels.", "Warning", wxOK | wxICON_INFORMATION, this);
					else {
						wxArrayInt selected;
						if (TrackChoices(selected, "File contains following tracks, please choose track(s) to import", "Import track", ld, this)) {
							bool ok = false;
							for (size_t i = 0; i < selected.Count(); i++) {
								if (!project.AddTrack(ld, selected.Item(i), this)) {
									wxLogError("Error selecting track from file (assertion failed)");
				 					delete ld;
								} else
									ok = true;
			 				}
							if (ok)
								Update();
						}
					}
					break;
				case IO_ERROR:
				case UNK_ERROR:
				case STRUCT_ERR:
					wxLogError("Error while opening file " + filename);
					break;
				case UNK_FORMAT:
					wxLogError("File format is unrecognized");
					break;
				case UNS_FORMAT:
					wxLogError("File format is recognized, but unsupported");
					break;
				default:
					wxLogError("Assertion failed while importing track(s)");
			}
			if (!closed)
				file->Close();
		} else
			wxLogError("Unable to open file " + filename + ", please check the permissions");
	}
}

void Frame::OnPlayBut(wxCommandEvent & event) {
	if (!project.Play())
		wxLogError("Unable to start playback");
}

void Frame::OnPauseBut(wxCommandEvent & event) {
	if (!project.Pause())
		wxLogError("Unable to pause/resume playback");
}

void Frame::OnStopBut(wxCommandEvent & event) {
	if (!project.Stop())
		wxLogError("Unable to stop playback");
}

void Frame::OnTimer(wxTimerEvent & event) {
	samples_t orig = project.GetCursor();
	samples_t updated = project.UpdateCursor();
	
	if (orig != updated) {

		wxClientDC dc(area);
		area->PrepareDC(dc);
	
		int width, height;
		area->GetVirtualSize(&width, &height);
	
		dc.BeginDrawing();
		Draw::XCursor(&dc);
		dc.SetLogicalFunction(wxINVERT);
	
		int x;
		int y;
	
		if (project.GetCursorVisual()) {
			x = orig;
			for (track_t i = 0; i < project.NumTracks(); i++) {
				y = (TRACK_HEIGHT + TRACK_SPACE) * i + TRACK_SPACE;
				dc.DrawLine(TRACK_LEFT + 1 + x, y + 1, TRACK_LEFT + 1 + x, y + 1 + (TRACK_HEIGHT - TRACK_RAW));
			}
		}
	
		x = updated;
		for (track_t i = 0; i < project.NumTracks(); i++) {
			y = (TRACK_HEIGHT + TRACK_SPACE) * i + TRACK_SPACE;
			dc.DrawLine(TRACK_LEFT + 1 + x, y + 1, TRACK_LEFT + 1 + x, y + 1 + (TRACK_HEIGHT - TRACK_RAW));
		}
	
		project.SetCursorVisual(true);
	
		dc.EndDrawing();
	}
}

void Frame::Update() {
	area->Update();
	controls->Update();
}

Area::Area(wxFrame * parent, Controls * controls) 
	: wxScrolledWindow(parent) {
	SetThemeEnabled(true);
	ctrl = controls;
}

void Area::OnPaint(wxPaintEvent & event) {
	wxPaintDC dc(this);
	PrepareDC(dc);
	
	int width, height;
	GetVirtualSize(&width, &height);
	
	for (track_t i = 0; i < project.NumTracks(); i++) {
		wxRect f;
		samples_t peaks = project.TrackPeaks(i);
		
		f.x = TRACK_LEFT;
		f.y = (TRACK_HEIGHT + TRACK_SPACE) * i + TRACK_SPACE;
		f.width = peaks + TRACK_RAW;
		f.height = TRACK_HEIGHT;
		
		Draw::Medium(&dc, false);
		dc.DrawRectangle(f);
		Draw::Bevel(dc, true, f);
		
		for (int x = 0; x < peaks; x++) {
			normsample_t max = (TRACK_HEIGHT - TRACK_RAW) / (normsample_t) 2;
			normsample_t amp = project.GetPeak(i, x) * (max / PEAK_MAX);

			int y1 = (int) (max - amp);
			int y2 = (int) (max + amp);
			
			Draw::Peak(&dc);
			dc.DrawLine(TRACK_LEFT + 1 + x, f.y + 1 + y1, TRACK_LEFT + 1 + x, f.y + 1 + y2);
		}
	}
}

void Area::OnScroll(wxScrollWinEvent & event) {
	int originx;
	int originy;

	GetViewStart(&originx, &originy);
	ctrl->CScroll(0, originy);
	
	event.Skip();
}

void Area::Update() {
	int vheight = (TRACK_HEIGHT + TRACK_SPACE) * project.NumTracks() + TRACK_SPACE;
	int vwidth = 0;
	
	for (track_t i = 0; i < project.NumTracks(); i++) {
		samples_t peaks = project.TrackPeaks(i);
		if (peaks > vwidth)
			vwidth = peaks;
	}
	
	SetScrollbars(1, 1, TRACK_LEFT + TRACK_RAW + vwidth + TRACK_RIGHT, vheight, 0, 0, false);
}

Controls::Controls(wxFrame * parent)
	: wxWindow(parent, -1, wxDefaultPosition, wxSize(CONTROLS, -1), 0) {
	
	origin = wxPoint(0, 0);
	boxes = NULL;
	mutes = NULL;
	SetThemeEnabled(true);
}

void Controls::CScroll(const int x, const int y) {
	ScrollWindow(origin.x - x, origin.y - y);
	origin = wxPoint(x, y);
}

void Controls::OnMute(wxCommandEvent & event) {
	wxButton * btn = (wxButton * ) event.GetEventObject();
	track_t track = event.GetId() - ID_FDynBut;

	if (project.GetMuteTrack(track)) {
		project.SetMuteTrack(track, false);
		btn->SetLabel("Mute");
	} else {
		project.SetMuteTrack(track, true);
		btn->SetLabel("Unmute");
	}
}

void Controls::Update() {
	track_t tracks = project.NumTracks();
	wxString tmp;

	if (mutes != NULL)
		delete[] mutes;
	if (boxes != NULL)
		delete[] boxes;

	if (tracks > 0) {
		boxes = new wxStaticBox[tracks];
		mutes = new wxButton[tracks];

		for (track_t i = 0; i < tracks; i++) {
			boxes[i].Create(this, -1, wxString("Track ") << (i + 1), wxPoint(TRACK_LEFT, (TRACK_HEIGHT + TRACK_SPACE) * i + TRACK_SPACE), wxSize(CONTROLS - TRACK_LEFT - TRACK_RIGHT, TRACK_HEIGHT), wxCAPTION);
			
			if (project.GetMuteTrack(i))
				tmp = "Unmute";
			else
				tmp = "Mute";
				
			mutes[i].Create(this, ID_FDynBut + i, tmp, wxPoint(TRACK_LEFT + 10, (TRACK_HEIGHT + TRACK_SPACE) * i + TRACK_SPACE + 20));
			
		}
		
		Connect(ID_FDynBut, ID_FDynBut + tracks - 1, wxEVT_COMMAND_BUTTON_CLICKED, (wxObjectEventFunction) (wxEventFunction) (wxCommandEventFunction) &Controls::OnMute);
	} else {
		boxes = NULL;
		mutes = NULL;
	}
}

