/***********************************************************************

  FinalMix: LoaderWAV.cpp
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  0.9.2		Initial

***********************************************************************/

#include "generic.h"
#include "LoaderWAV.h"

#include <wx/filename.h>
#include <wx/listimpl.cpp>

WX_DEFINE_LIST(Datas);

typedef struct {
	wxUint16 Format;
	wxUint16 Channels;
	wxUint32 SampleRate;
	wxUint32 BytesPerSec;
	wxUint16 BlockAlign;
	wxUint16 BitWidth;
} Wave_PCM_header;

typedef struct {
	wxUint32 esf_rec1;
	wxUint16 esf_rec2;
	wxUint16 esf_rec3;
	char esf_field[8];
} Esf_subformat;

typedef struct {
	wxUint16 Format;
	wxUint16 Channels;
	wxUint32 SampleRate;
	wxUint32 BytesPerSec;
	wxUint16 BlockAlign;
	wxUint16 BitWidth;
	wxUint16 ExtraBytes;
	wxUint16 ValidBits;
	wxUint32 ChannelMask;
	Esf_subformat esf;
} Wave_PCM_extensible_header;

typedef union {
	wxUint16 FormatTag;
	Wave_PCM_header PCM;
	Wave_PCM_extensible_header PCMext;
	char padding[512];
} Wave_header;

typedef union {
	wxUint32 id;
	char chars[4];
} fourcc;

typedef struct {
	fourcc tag;
	wxUint32 len;
} chunk_header;

inline bool cmp_fourcc(const fourcc cc, const char * str) {
	return(strncmp((const char * ) &cc.chars, str, 4) == 0);
}

enum {
	WAVE_FORMAT_UNKNOWN = 0x0000,
	WAVE_FORMAT_PCM = 0x0001
};

WAVLoader::WAVLoader() {
	cf = NULL;
}

void WAVLoader::done() {
	if (cf != NULL) {
		if (cf->IsOpened())
			cf->Close();
		delete cf;
	}
	cf = NULL;
}

WAVLoader::~WAVLoader() {
	done();
}

wxString WAVLoader::Name() const {
	return(wxString("RIFF WAVE loader"));
}

enum status_t WAVLoader::Attach(wxFile * file, const wxString & name, const wxDateTime modtime) {
	done();
	if (file->IsOpened()) {
		fourcc tmp;
		chunk_header htmp;
		index_t state = 0;
		format_t fmt;
		track_t chans = 0;
		Datas tdatas;
		length_t bytes = 0;
		
		if (file->Seek(0) == wxInvalidOffset)
			return(IO_ERROR);
		
		if (file->Read(&htmp, sizeof(htmp)) != sizeof(htmp))
			return(IO_ERROR);
		if (!cmp_fourcc(htmp.tag, "RIFF"))
			return(UNK_FORMAT);
		
		if (file->Read(&tmp, sizeof(tmp)) != sizeof(tmp))
			return(IO_ERROR);
		if (!cmp_fourcc(tmp, "WAVE"))
			return(UNK_FORMAT);
		
		while (!file->Eof()) {
			if (file->Read(&htmp, sizeof(htmp)) != sizeof(htmp))
				return(IO_ERROR);
			
			if (cmp_fourcc(htmp.tag, "fmt ")) {
				Wave_header header;
				off_t tr;
				off_t rd;
				
				if (state & 1)
					return(STRUCT_ERR);
				
				tr = min(htmp.len, sizeof(header));
				rd = file->Read(&header, tr);
				
				if (rd != tr)
					return(IO_ERROR);
				
				switch (header.FormatTag) {
					case WAVE_FORMAT_PCM:
						if (rd < sizeof(Wave_PCM_header))
							return(STRUCT_ERR);
						if (header.PCM.BitWidth % 8 != 0)
							return(UNS_FORMAT);
						if ((header.PCM.BitWidth != 8) && (header.PCM.BitWidth != 16))
							return(UNS_FORMAT);
						fmt.encoding = PCM;
						chans = header.PCM.Channels;
						fmt.rate = header.PCM.SampleRate;
						switch (header.PCM.BitWidth) {
							case 8:
								fmt.samplewidth = BIT8;
								fmt.bitencoding = UNSIGNED;
								break;
							case 16:
								fmt.samplewidth = BIT16;
								fmt.bitencoding = SIGNED;
								break;
							default:
								return(UNS_FORMAT);
						}
						break;
					default:
						return(UNS_FORMAT);
				}
				
				if (htmp.len > rd)
					file->Seek(htmp.len - rd, wxFromCurrent);
				state |= 1;
			} else if (cmp_fourcc(htmp.tag, "data")) {
				DataPos * data = new DataPos;
				
				data->pos = file->Tell();
				data->len = htmp.len;
				bytes += htmp.len;
				tdatas.Append(data);
				file->Seek(htmp.len, wxFromCurrent);
				state |= 2;
			}
		}
		
		if (state & 3) {
			if ((fmt.samplewidth == 0) || (chans == 0))
				return(STRUCT_ERR);
			format = fmt;
			switch (fmt.samplewidth) {
				case BIT8:
					format.samples = bytes / chans;
					break;
				case BIT16:
					format.samples = bytes / 2 / chans;
					break;
				default:
					FATAL("Unreachable code");
					break;
			}
			tracks = chans;
			datas = tdatas;
		}
		
		cf = file;
		fname = name;
		mod_time = modtime;
		return(OK);
	} else
		return(IO_ERROR);
}

track_t WAVLoader::NumIntTracks() const {
	if (cf != NULL)
		return tracks;
	else
		return 0;
}

wxString WAVLoader::FileName() const {
	if (cf != NULL)
		return fname;
	else
		return "";
}

wxDateTime WAVLoader::GetModTime() const {
	return mod_time;
}

length_t WAVLoader::GetRawLength() const {
	return cf->Length();
}

const format_t * WAVLoader::TrackFormat(const track_t itrack) const {
	if ((cf != NULL) && (itrack < tracks))
		return &format;
	else
		return NULL;
}

samples_t WAVLoader::GetSamples(const track_t itrack, const samples_t pos, const samples_t len, void * data) {
	samples_t offset = pos * tracks;
	samples_t size = len * tracks;
	samples_t done = 0;
	wxDatasNode * node = NULL;
	DataPos * datapos = NULL;
	void * tmp = NULL;
	void * sample = AllocSamples(format, 1);
	
	if ((!sample) || (itrack >= tracks))
		return(0);
			
	for (node = datas.GetFirst(); node != NULL; node = node->GetNext()) {
		datapos = node->GetData();
		
		if (datapos->len <= offset * SampleLen(format))
			offset -= datapos->len / SampleLen(format);
		else
			break;
	}
	
	if ((node) && (datapos)) {
		if (tracks > 1) {
			tmp = AllocSamples(format, size);
			if (!tmp)
				return(0);
		}
		while (size > 0) {
			cf->Seek(datapos->pos + offset * SampleLen(format));
			size_t tr = min(size * SampleLen(format), datapos->len - offset);
			size_t rd;
			if (tracks > 1)
				rd = cf->Read(GetSample(format, done, tmp), tr);
			else
				rd = cf->Read(GetSample(format, done, data), tr);
			if (rd != tr) {
				if (tmp)
					FreeSamples(format, tmp);
				return(0);
			}
			size -= tr / SampleLen(format);
			done += tr / SampleLen(format);
			offset = 0;
			node->GetNext();
			if (!node)
				break;
			datapos = node->GetData();
			if (!datapos)
				break;
		}
		if (size != 0) {
			if (tmp)
				FreeSamples(format, tmp);
			return(0);
		}
		if (tmp) {
			for (samples_t i = 0; i < len; i++) {
				memcpy(sample, GetSample(format, i * tracks + itrack, tmp), SampleLen(format));
				memcpy(GetSample(format, i, data),  sample, SampleLen(format));
			}
			FreeSamples(format, tmp);
		}
		return(len);
	} else
		return(0);
}

Loader * WAVLoader::RegisterLoader() {
	return(new WAVLoader);
}
