/***********************************************************************

  FinalMix: Peaks.cpp
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  0.9.2		Initial
  0.9.3		Really supports peaks

***********************************************************************/

#include "generic.h"
#include "Peaks.h"
#include <wx/progdlg.h>
#include <wx/msgdlg.h>

Peaks::Peaks(Loader * ld, const track_t itrack, const wxString & name, const ratio_t density) {
	ASSERT(ld != NULL);

	const format_t * _fmt = ld->TrackFormat(itrack);
	ASSERT(_fmt != NULL);
		
	format_t fmt = *_fmt;
	loader = ld;
	track = itrack;
	ratio = density;
	
	if (wxFile::Exists(name)) {
		if (!wxFile::Access(name, wxFile::read_write)) {
			// FIXME: indicate reason
			peaks = NULL;
		}
		peaks = new wxFile(name, wxFile::read_write);
		if (!check_header(loader->GetModTime(), loader->GetRawLength()))
			empty = true;
		else
			empty = false;
	} else
		empty = true;
	
	if (empty) {	
		peaks = new wxFile;
		if (peaks) {
			peaks->Create(name, true);
			if (peaks->IsOpened()) {
				if (!write_header(loader->GetModTime(), loader->GetRawLength()))
					not_persistent();
				peaks->Close();
				peaks->Open(name, wxFile::read_write);
			} else {
				// FIXME: indicate reason
				peaks = NULL;
			}
		}
	}
}

bool Peaks::Synchronize(wxWindow * parent) {
	bool retval = true;

	if (!peaks)
		return true;
		
	if (empty) { 
		wxProgressDialog pdlg("Opening", "Updating peak file", FULL_PERCENT, parent, wxPD_CAN_ABORT | wxPD_ELAPSED_TIME | wxPD_ESTIMATED_TIME | wxPD_REMAINING_TIME | wxPD_AUTO_HIDE);
			
		const format_t * fmt = loader->TrackFormat(track);
		void * rbuf = AllocSamples(*fmt, ratio);
	
		if (rbuf) {
			samples_t left = fmt->samples;
			samples_t pos = 0;
		
			peaks->Seek(datapos, wxFromStart);
	
			while (left > 0) {
				samples_t tr = min(left, (samples_t) ratio);
				samples_t rd = loader->GetSamples(track, pos, tr, rbuf);
				if (rd > 0) {
					bool ret;
					normsample_t out;
					size_t wr;
				
					ret = SamplesMaxNorm(*fmt, rd, rbuf, &out, PEAK_MAX);
					ASSERT(ret);
				
					wr = peaks->Write((const void * ) &out, sizeof(out));
					if (wr != sizeof(out)) {
						// FIXME: indicate error visually
						retval = false;
						not_persistent();
						break;
					}
					
					left -= rd;
					pos += rd;
					normsample_t percent = (samples_ex_t) pos / (normsample_t) (samples_ex_t) fmt->samples * FULL_PERCENT;
					if (!pdlg.Update((int) percent)) {
						wxMessageDialog mdlg(parent, "Cancel updating peak file?", "Peak file update", wxYES_NO | wxNO_DEFAULT | wxICON_QUESTION);
						if (mdlg.ShowModal() == wxID_YES) {
							retval = false;
							not_persistent();
							break;
						} else
							pdlg.Resume();
					}
				}
			}
			FreeSamples(*fmt, rbuf);
		}
	}
	
	return retval;
}

Peaks::~Peaks() {
	if (peaks) {
		peaks->Close();
		delete peaks;
	}
}

bool Peaks::GetPeak(const samples_t ppos, const ratio_t density, normsample_t * peak) {
	// FIXME: check for ranges

	if (peaks) {	/* with persistent file */
		if (density == ratio) {
			peaks->Seek(datapos + ppos * sizeof(normsample_t), wxFromStart);
			off_t rd = peaks->Read((void * ) peak, sizeof(normsample_t));
			if (rd != sizeof(normsample_t))
				return false;
		} else {
			FATAL("TODO");		// FIXME
		}
	} else {		/* w/o persistent file */
		const format_t * fmt = loader->TrackFormat(track);
		void * rbuf = AllocSamples(*fmt, ratio);
	
		if (rbuf) {
			samples_t tr = min(fmt->samples - ppos * ratio, (samples_t) ratio);
			samples_t rd = loader->GetSamples(track, ppos * ratio, tr, rbuf);
			if (rd > 0) {
				bool ret;
				
				ret = SamplesMaxNorm(*fmt, rd, rbuf, peak, PEAK_MAX);
				ASSERT(ret);
			}
		}
		FreeSamples(*fmt, rbuf);
	}
	return true;
}

bool Peaks::write_header(const wxDateTime modtime, const length_t length) {
	if (!write_tag(FM_PEAK_TAG))
		return false;
	
	size_t ret;
	
	ret = peaks->Write(&fmfile_version_major, sizeof(version_t));
	if (ret != sizeof(version_t))
		return false;
		
	ret = peaks->Write(&fmfile_version_minor, sizeof(version_t));
	if (ret != sizeof(version_t))
		return false;
		
	stamppos = peaks->Tell();
		
	ret = peaks->Write(&ratio, sizeof(ratio_t));
	if (ret != sizeof(ratio_t))
		return false;
		
	ret = peaks->Write(&length, sizeof(length_t));
	if (ret != sizeof(length_t))
		return false;
		
	length_t ticks = modtime.GetTicks();
	ret = peaks->Write(&ticks, sizeof(length_t));
	if (ret != sizeof(length_t))
		return false;
		
	// FIXME write number of peaks (reserve space)
		
	datapos = peaks->Tell();
		
	return true;
}

bool Peaks::check_header(const wxDateTime modtime, const length_t length) {
	peaks->Seek(0, wxFromStart);
	
	size_t tag_len = strlen(FM_PEAK_TAG);
	char * tag = new char[tag_len + 1];
	size_t ret;
	
	if ((!read_tag(tag, tag_len)) || (strcmp(tag, FM_PEAK_TAG) != 0))
		return false;
		
	delete[] tag;
	
	version_t major;
	version_t minor;
	
	ret = peaks->Read(&major, sizeof(version_t));
	if ((ret != sizeof(version_t)) || (major != fmfile_version_major))
		return false;
		
	ret = peaks->Read(&minor, sizeof(version_t));
	if ((ret != sizeof(version_t)) || (minor != fmfile_version_minor))
		return false;
		
	stamppos = peaks->Tell();
	
	ratio_t rratio;
	length_t llength;
	length_t tticks;
	
	ret = peaks->Read(&rratio, sizeof(ratio_t));
	if ((ret != sizeof(ratio_t)) || (rratio != ratio))
		return false;
		
	ret = peaks->Write(&llength, sizeof(length_t));
	if ((ret != sizeof(length_t)) || (llength != length))
		return false;
		
	ret = peaks->Read(&tticks, sizeof(length_t));
	if ((ret != sizeof(length_t)) || (tticks != modtime.GetTicks()))
		return false;
		
	datapos = peaks->Tell();

	return true;	
}

bool Peaks::write_tag(const char * str) {
	off_t len = strlen(str);
	size_t ret;
	
	ret = peaks->Write(str, sizeof(char) * len);
	if (ret != sizeof(char) * len)
		return false ;
	
	return true;
}

bool Peaks::read_tag(char * str, size_t size) {
	size_t ret;

	ret = peaks->Read(str, sizeof(char) * size);
	if (ret != sizeof(char) * size)
		return false;
		
	str[size] = 0;
	
	return true;
}

bool Peaks::write_string(const char * str) {
	wxUint64 len = strlen(str);
	size_t ret;
	
	ret = peaks->Write(&len, sizeof(len));
	if (ret != sizeof(len))
		return false;
	
	ret = peaks->Write(str, sizeof(char) * len);
	if (ret != sizeof(char) * len)
		return false;
	
	return true;
}

void Peaks::not_persistent() {
	if (peaks) {
		if (peaks->IsOpened())
			peaks->Close();
		delete peaks;
		peaks = NULL;
	}
}
