/***********************************************************************

  FinalMix: Project.cpp
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  0.9.2		Initial
  0.9.3		Cleanup

***********************************************************************/

#include "generic.h"
#include "Project.h"

#include <wx/listimpl.cpp>

Project project(DEF_PEAK_DIR, DEF_PEAK_RATIO);

WX_DEFINE_LIST(EditCuts);
WX_DEFINE_LIST(Tracks);
WX_DEFINE_LIST(Loaders);
WX_DEFINE_LIST(Devices);

struct slist_t {
	void * data;
	samples_t samples;
	bool mute;
	struct slist_t * next;
};

PlayCallback::PlayCallback() {
	// TODO
}

PlayCallback::~PlayCallback() {
	// TODO
}

samples_t PlayCallback::OnBuffer(void * buf, const samples_t len, const samples_t pos, const format_ex_t & format) {
	wxMutexLocker lock(project.track_mutex);

	struct slist_t * tmp = NULL; 
	struct slist_t * ac = NULL;
	format_t ofmt;
	
	FormatAlias(format, ofmt);
	
	for (index_t i = 0; i < project.tracks.GetCount(); i++) {
		wxTracksNode * node = project.tracks.Item(i);
		track_desc_t * trck;
	
		if (node) {
			trck = node->GetData();
			if (trck) {
				if (!trck->mute) {
					const format_t * fmt = trck->loader->TrackFormat(trck->itrack);
					samples_t nlen = SamplesCmpIn(*fmt, ofmt, len);
					samples_t npos = SamplesCmpIn(*fmt, ofmt, pos);
				
					void * data = AllocSamples(*fmt, nlen);
				
					if (data) {
						void * dout = AllocSamples(format, len);
					
						if (dout) {
							samples_t rd = trck->loader->GetSamples(trck->itrack, npos, nlen, data);
							samples_t dn = SamplesConv(*fmt, format, rd, data, dout);
						
							if (!tmp) {
								tmp = new struct slist_t;
								if (tmp) {
									tmp->data = dout;
									tmp->samples = dn;
									tmp->next = NULL;
								}
								ac = tmp;
							} else {
								ac->next = new struct slist_t;
								if (ac->next) {
									ac->next->data = dout;
									ac->next->samples = dn;
									ac->next->next = NULL;
								}
								ac = ac->next;
							}
						}
						FreeSamples(*fmt, data);
					}
				}
			} else
				break;
		} else
			break;
	}
	
	ClearSamples(format, len, buf);

	for (ac = tmp; ac != NULL; ac = ac->next)
		AddSamples(format, ac->samples, len, ac->data, buf);
		
	return len;
}

Project::Project(const wxString & peakdir, const ratio_t ratio) {
	pdev = NULL;
	rdev = NULL;
	pdir = peakdir; // configure
	pratio = ratio; // configure
	pformat.encoding = PCM; // configure
	pformat.bitencoding = SIGNED;
	pformat.samplewidth = BIT16;
	pformat.rate = 44100;
	pformat.channels = 2;
	changed = false;
}

bool Project::AddTrack(Loader * ld, const track_t itrack, wxWindow * parent) {

	// FIXME check if not playing (with mutex)

	const format_t * fmt = ld->TrackFormat(itrack);

	if ((ld) && (fmt)) {
		track_desc_t * trck = new track_desc_t;
		
		if (trck) {
			wxString suffix;
			
			suffix << "-" << itrack;
			Peaks * pks = new Peaks(ld, itrack, SetFilename(pdir, ld->FileName(), FM_PEAK_EXT, suffix), pratio);
			if (pks) {
				if (pks->Synchronize(parent)) {
					trck->loader = ld;
					trck->itrack = itrack;
					trck->format = *fmt;
					trck->peaks = pks;
					trck->mute = false;
					
					tracks.Append(trck);
					
					changed = true;
					return true;
				}
				delete pks;
			}
		}
	}
	return false;
}

bool Project::SetMuteTrack(const track_t i, const bool state) {
	wxMutexLocker lock(track_mutex);

	wxTracksNode * node = project.tracks.Item(i);
	track_desc_t * trck;
	
	if (node) {
		trck = node->GetData();
		if (trck) {
			trck->mute = state;
			return true;
		}
	}
	
	return false;
}

bool Project::GetMuteTrack(const track_t i) {
	wxMutexLocker lock(track_mutex);

	wxTracksNode * node = project.tracks.Item(i);
	track_desc_t * trck;
	
	if (node) {
		trck = node->GetData();
		if (trck)
			return trck->mute;
	}
	
	return true;
}

track_t Project::NumTracks() {
	wxMutexLocker lock(track_mutex);

	return tracks.GetCount();
}

samples_t Project::TrackPeaks(const track_t track) {
	wxMutexLocker lock(track_mutex);

	wxTracksNode * node = tracks.Item(track);
	track_desc_t * trck;
	
	if (node) {
		trck = node->GetData();
		if (trck) {
			const format_t * fmt = trck->loader->TrackFormat(trck->itrack);
			if (fmt) {
				samples_t peaks = fmt->samples / pratio;
				if (fmt->samples % pratio != 0)
					peaks++;
					
				return peaks;
			}
		}
	}
	return 0;
}

normsample_t Project::GetPeak(const track_t track, const samples_t ppos) {
	wxMutexLocker lock(track_mutex);

	wxTracksNode * node = tracks.Item(track);
	track_desc_t * trck;
	normsample_t ret = 0;
	
	if (node) {
		trck = node->GetData();
		if (trck) {
			const format_t * fmt = trck->loader->TrackFormat(trck->itrack);
			
			if (fmt) {
				if (!trck->peaks->GetPeak(ppos, pratio, &ret))
					ret = 0;
			}
		}
	}
	
	return ret;
}

bool Project::RegisterLoader(Loader * ( * reg)()) {
	loader_t * rec = new loader_t;
	
	if (rec) {
		rec->reg = reg;
		loaders.Append(rec);
		return true;
	}
	return false;
}

Loader * Project::GetLoader(const index_t index) const {
	wxLoadersNode * node = loaders.Item(index);
	loader_t * rec;
	
	if (node) {
		rec = node->GetData();
		if (rec)
			return rec->reg();
	}
	return NULL;
}

bool Project::RegisterDevice(Device * ( * reg)()) {
	device_t * rec = new device_t;
	
	if (rec) {
		rec->reg = reg;
		devices.Append(rec);
		return true;
	}
	return false;
}

Device * Project::GetDevice(const index_t index) const {
	wxDevicesNode * node = devices.Item(index);
	device_t * rec;
	
	if (node) {
		rec = node->GetData();
		if (rec)
			return rec->reg();
	}
	return NULL;
}

bool Project::Changed() const {
	return changed;
}

bool Project::Play() {
	if (pdev) {
		if (pdev->Init(pformat, pcbk))
			return pdev->Play();
	}
	return false;
}

bool Project::Pause() {
	if (pdev) {
		return pdev->Pause();
	}
	return false;
}

bool Project::Stop() {
	if (pdev) {
		return pdev->Stop();
	}
	return false;
}

bool Project::GetCursorVisual() const {
	return cursor_shown;
}

void Project::SetCursorVisual(bool shown) {
	cursor_shown = shown;
}

samples_t Project::GetCursor() const {
	return (cursor / pratio);
}

samples_t Project::UpdateCursor() {
	if (pdev)
		cursor = pdev->GetPosition();
		
	return (cursor / pratio);
}

bool Project::SetPlayDevice(Device * dev) {
	if (pdev)
		pdev->Done();

	if (dev) {
		pdev = dev;
		return true;
	}
	return false;
}
