/***********************************************************************

  FinalMix: Project.h
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  0.9.2		Managing project
  0.9.3		Cleanup

**********************************************************************/

#ifndef _FM_PROJECT_H_
#define _FM_PROJECT_H_

#include "generic.h"
#include "Peaks.h"
#include "Loaders.h"
#include "Devices.h"
#include <wx/wx.h>
#include <wx/list.h>
#include <wx/thread.h>

/** Cut kind type.
 *
 */

enum cut_t {
	CUT_DELETE = 1											/**< Delete cut */
};

/** Track cut type.
 *
 */

typedef struct {
	enum cut_t type;										/**< Cut kind */
	samples_t start;										/**< Start of the cut in samples */
	samples_t len;											/**< Length of the cut in samples */
} editcut_t;

/** List of cuts in the track type.
 *
 */

WX_DECLARE_LIST(editcut_t, EditCuts);

/** Track description type.
 *
 */

typedef struct {
	Loader * loader;										/**< Loader used for this track */
	track_t itrack;											/**< Index of the track as seen to the loader */
	Peaks * peaks;											/**< Peak generator for this track */
	format_t format;										/**< Format description */
	bool mute;												/**< Track is muted */
	EditCuts cuts;											/**< Cuts */
} track_desc_t;

/** List of tracks in the project type.
 *
 */

WX_DECLARE_LIST(track_desc_t, Tracks);

/** Loader type.
 *
 */

typedef struct {
	Loader * ( * reg)();									/**< Registrator for the loader */
	int tag;												/**< Local value */
} loader_t;

/** List of loaders type.
 *
 */

WX_DECLARE_LIST(loader_t, Loaders);

/** Device type.
 *
 */

typedef struct {
	Device * ( * reg)();									/**< Registrator for the device */
	int tag;												/**< Local value */
} device_t;

/** List of devices type.
 *
 */

WX_DECLARE_LIST(device_t, Devices);

/** Playback Callback class.
 *
 * This class is used by the playback device to give the main process an acknowledge
 * to generate (mix) next part of the output buffer.
 *
 */

class PlayCallback : public DeviceCallback {
	public:
		PlayCallback();										/**< Constructor */
		virtual ~PlayCallback();							/**< Destructor */
		
		/** OnBuffer event.
		 *
		 * This event method is executed when the playback device demands to mix next
		 * part of the output buffer.
		 *
		 * @param buf    Output (mix buffer).
		 * @param len    Number of samples to mix.
		 * @param pos    Absolute position in the playback stream where to mix the first sample.
		 * @param format Output format description to generate.
		 *
		 */
		
		virtual samples_t OnBuffer(void * buf, const samples_t len, const samples_t pos, const format_ex_t & format);
	private:
};

/** Project class.
 *
 * Base class of a project. Project contains tracks, cuts, loaders, devices etc.
 *
 */

class Project {
	public:
	
		/** Constructor.
		 *
		 * Creates new project.
		 *
		 * @param peakdir Directory for storing peak files.
		 * @param ratio   Initial peak display ratio.
		 *
		 */
	
		Project(const wxString & peakdir, const ratio_t ratio);
		
		/** Add new track to the project.
		 *
		 * Adds new track (via its loader) to the project.
		 *
		 * @param ld     Loader to add.
		 * @param itrack Which subtrack in the loader to use.
		 * @param parent wxWindow to use for showing visual feedback (progress bars etc.).
		 *
		 */
		
		bool AddTrack(Loader * ld, const track_t itrack, wxWindow * patent);
		
		/** Remove track from project.
		 *
		 * Removes track from the project.
		 *
		 * @param track  Index of the track in the project.
		 *
		 */
		
		bool RemoveTrack(const track_t track);
		
		bool SetMuteTrack(const track_t track, const bool state);
		bool GetMuteTrack(const track_t track);
		
		track_t NumTracks();										/**< Number of tracks */
		samples_t TrackPeaks(const track_t track);					/**< FIXME */
		normsample_t GetPeak(const track_t track, const samples_t ppos);	/**< FIXME */
		
		bool Changed() const;										/**< Is the project changed since last save/revert? */
		
		bool Play();												/**< Play the project */
		bool Pause();												/**< Pause playback */
		bool Stop();												/**< Stop playback */

		bool GetCursorVisual() const;
		void SetCursorVisual(bool shown);
		samples_t GetCursor() const;
		samples_t UpdateCursor();
		
		/** Register new loader class.
		 *
		 * Creates needed instances of new loader class using loader's register method.
		 *
		 * @param reg Loader's register method.
		 *
		 */
		
		bool RegisterLoader(Loader * ( * reg)());
		index_t NumLoaders() const;									/**< Number of loaders */
		Loader * GetLoader(const index_t index) const;				/**< FIXME */
		
		/** Register new device class.
		 *
		 * Creates needed instances of new loader class using device's register method.
		 *
		 * @param reg Device's register method.
		 *
		 */
		
		bool RegisterDevice(Device * ( * reg)());
		index_t NumDevices() const;									/**< Number of devices */
		Device * GetDevice(const index_t index) const;				/**< FIXME */
		
		/** Set current playback device.
		 *
		 * Sets a device to be used as playback device for the project.
		 *
		 * @param dev Playback device to be used as audio output for the project.
		 *
		 */
		
		bool SetPlayDevice(Device * dev);
		
		friend samples_t PlayCallback::OnBuffer(void * buf, const samples_t len, const samples_t pos, const format_ex_t & format);	/**< Playback Callback can access project's private structures */
	private:
		format_ex_t pformat;										/**< Output format of the project */
		Tracks tracks;												/**< Tracks list */
		wxMutex track_mutex;
		Loaders loaders;											/**< Loaders list */
		Devices devices;											/**< Devices list */
		
		PlayCallback pcbk;											/**< Playback callback */
		
		ratio_t pratio;												/**< Peak display ratio */
		
		wxString pdir;												/**< Peak file directory */
		
		Device * pdev;												/**< Playback device */
		Device * rdev;												/**< Record device */
			
		bool changed;												/**< Changed since last save/revert */
		bool cursor_shown;
		samples_t cursor;
};

extern Project project;

#endif
