/***********************************************************************

  FinalMix: generic.cpp
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  0.9.3		Initial, generic functions

***********************************************************************/

#include "generic.h"
#include <wx/filename.h>

const wxUint16 fmfile_version_major = 1;
const wxUint16 fmfile_version_minor = 0;


const sample_pcm8u_t sample_pcm8u_min = 0;
const sample_pcm8u_t sample_pcm8u_max = 255;
const sample_pcm8u_t sample_pcm8u_0 = 127;
const sample_pcm8u_t sample_pcm8u_dr = 128;

const sample_pcm8s_t sample_pcm8s_min = -128;
const sample_pcm8s_t sample_pcm8s_max = 127;
const sample_pcm8s_t sample_pcm8s_0 = 0;
const sample_pcm8s_t sample_pcm8s_dr = 128;

const sample_pcm16u_t sample_pcm16u_min = 0;
const sample_pcm16u_t sample_pcm16u_max = 65535;
const sample_pcm16u_t sample_pcm16u_0 = 32767;
const sample_pcm16u_t sample_pcm16u_dr = 32768;

const sample_pcm16s_t sample_pcm16s_min = -32768;
const sample_pcm16s_t sample_pcm16s_max = 32767;
const sample_pcm16s_t sample_pcm16s_0 = 0;
const sample_pcm16s_t sample_pcm16s_dr = 32768;

void Fatal(const char * format, ...) {
	va_list pars;

	fprintf(stderr, "%s INTERNAL FATAL ERROR\n ", FM_NAME_UPPER);
	
	va_start(pars, format);
	vfprintf(stderr, format, pars);
	va_end(pars);
	
	fprintf(stderr, "\n");
	
#ifdef DEBUG
	exit(-1);
#endif
}

wxString SetFilename(const wxString path, const wxString name, const wxString ext, const wxString fsuffix) {
	wxString pth;
	wxString fn;
	wxString ex;
	
	wxFileName::SplitPath(name, &pth, &fn, &ex);							/* split the original filename into components */
	
	wxFileName tmp(path, fn + fsuffix, ext);								/* creates the new filename according the template: $path$fn$fssuffix$ext */
	
	return tmp.GetFullPath();
}

size_t sample_len_generic(const enum encoding_t & encoding, const enum bitencoding_t & bitencoding, const enum samplewidth_t & samplewidth) {
	switch (encoding) {
		case PCM:
			switch (bitencoding) {
				case UNSIGNED:
					switch (samplewidth) {
						case BIT8:
							return sizeof(sample_pcm8u_t);
						case BIT16:
							return sizeof(sample_pcm16u_t);
						default:
							FATAL("Undefined format.samplewidth");
							return 0;
					}
				case SIGNED:
					switch (samplewidth) {
						case BIT8:
							return sizeof(sample_pcm8s_t);
						case BIT16:
							return sizeof(sample_pcm16s_t);
						default:
							FATAL("Undefined format.samplewidth");
							return 0;
					}
				default:
					FATAL("Undefined format.bitencoding");
					return 0;
			}
		default:
			FATAL("Undefined format.encoding");
			return 0;
	}
}

size_t SampleLen(const format_t & format) {
	return sample_len_generic(format.encoding, format.bitencoding, format.samplewidth);
}

size_t SampleLen(const format_ex_t & format) {
	return sample_len_generic(format.encoding, format.bitencoding, format.samplewidth);
}

void * alloc_samples_generic(const enum encoding_t & encoding, const enum bitencoding_t & bitencoding, const enum samplewidth_t & samplewidth, const samples_t count, const index_t channels) {
	switch (encoding) {
		case PCM:
			switch (bitencoding) {
				case UNSIGNED:
					switch (samplewidth) {
						case BIT8:
							return (new sample_pcm8u_t[count * channels]);
						case BIT16:
							return (new sample_pcm16u_t[count * channels]);
						default:
							FATAL("Undefined format.samplewidth");
							return NULL;
					}
				case SIGNED:
					switch (samplewidth) {
						case BIT8:
							return (new sample_pcm8s_t[count * channels]);
						case BIT16:
							return (new sample_pcm16s_t[count * channels]);
						default:
							FATAL("Undefined format.samplewidth");
							return NULL;
					}
				default:
					FATAL("Undefined format.bitencoding");
					return NULL;
			}
		default:
			FATAL("Undefined format.encoding");
			return NULL;
	}
}

void * AllocSamples(const format_t & format, const samples_t count) {
	return alloc_samples_generic(format.encoding, format.bitencoding, format.samplewidth, count, 1);
}

void * AllocSamples(const format_ex_t & format, const samples_t count) {
	return alloc_samples_generic(format.encoding, format.bitencoding, format.samplewidth, count, format.channels);
}

bool FreeSamples(const format_t & format, const void * data) {
	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							delete[] (sample_pcm8u_t * ) data;
							break;
						case BIT16:
							delete[] (sample_pcm16u_t * ) data;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							return false;
					}
					break;
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							delete[] (sample_pcm8s_t * ) data;
							break;
						case BIT16:
							delete[] (sample_pcm16s_t * ) data;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							return false;
					}
					break;
				default:
					FATAL("Undefined format.bitencoding");
					return false;
			}
			break;
		default:
			FATAL("Undefined format.encoding");
			return false;
	}
	
	return true;
}

bool samples_average_generic(const format_t & format, const samples_t samples, const void * inbuf, normsample_t * average) {
	normsample_t ac;
	size_t i;
	
	if (samples == 0)
		return false;

	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							ac = sample_pcm8u_0;
							for (i = 0; i < samples; i++) {
								ac += *((sample_pcm8u_t * ) inbuf + i) / (normsample_t) (samples_ex_t) samples;
							}
							*average = ac;
							break;
						case BIT16:
							ac = sample_pcm16u_0;
							for (i = 0; i < samples; i++) {
								ac += *((sample_pcm16u_t * ) inbuf + i) / (normsample_t) (samples_ex_t) samples;
							}
							*average = ac;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							return false;
					}
					break;
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							ac = sample_pcm8s_0;
							for (i = 0; i < samples; i++) {
								ac += *((sample_pcm8s_t * ) inbuf + i) / (normsample_t) (samples_ex_t) samples;
							}
							*average = ac;
							break;
						case BIT16:
							ac = sample_pcm16s_0;
							for (i = 0; i < samples; i++) {
								ac += *((sample_pcm16s_t * ) inbuf + i) / (normsample_t) (samples_ex_t) samples;
							}
							*average = ac;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							return false;
					}
					break;
				default:
					FATAL("Undefined format.bitencoding");
					return false;
			}
			break;
		default:
			FATAL("Undefined format.encoding");
			return false;
	}
	return true;
}

bool samples_max_generic(const format_t & format, const samples_t samples, const void * inbuf, normsample_t * average) {
	normsample_t ac, cur;
	size_t i;
	
	if (samples == 0)
		return false;

	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							ac = sample_pcm8u_0;
							for (i = 0; i < samples; i++) {
								cur = abs(*((sample_pcm8u_t * ) inbuf + i));
								if (cur > ac)
									ac = cur;
							}
							*average = ac;
							break;
						case BIT16:
							ac = sample_pcm16u_0;
							for (i = 0; i < samples; i++) {
								cur = abs(*((sample_pcm16u_t * ) inbuf + i));
								if (cur > ac)
									ac = cur;
							}
							*average = ac;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							return false;
					}
					break;
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							ac = sample_pcm8s_0;
							for (i = 0; i < samples; i++) {
								cur = abs(*((sample_pcm8s_t * ) inbuf + i));
								if (cur > ac)
									ac = cur;
							}
							*average = ac;
							break;
						case BIT16:
							ac = sample_pcm16s_0;
							for (i = 0; i < samples; i++) {
								cur = abs(*((sample_pcm16s_t * ) inbuf + i));
								if (cur > ac)
									ac = cur;
							}
							*average = ac;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							return false;
					}
					break;
				default:
					FATAL("Undefined format.bitencoding");
					return false;
			}
			break;
		default:
			FATAL("Undefined format.encoding");
			return false;
	}
	return true;
}

bool SamplesAverage(const format_t & format, const samples_t samples, const void * inbuf, const void * outbuf) {
	bool retval;
	normsample_t ac;
	
	retval = samples_average_generic(format, samples, inbuf, &ac);

	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							*((sample_pcm8u_t *) outbuf) = (sample_pcm8u_t) ac;
							break;
						case BIT16:
							*((sample_pcm16u_t *) outbuf) = (sample_pcm16u_t) ac;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							retval = false;
							break;
					}
					break;
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							*((sample_pcm8s_t *) outbuf) = (sample_pcm8s_t) ac;
							break;
						case BIT16:
							*((sample_pcm16s_t *) outbuf) = (sample_pcm16s_t) ac;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							retval = false;
							break;
					}
					break;
				default:
					FATAL("Undefined format.bitencoding");
					retval = false;
					break;
			}
			break;
		default:
			FATAL("Undefined format.encoding");
			retval = false;
			break;
	}
	return retval;
}

bool SamplesMax(const format_t & format, const samples_t samples, const void * inbuf, const void * outbuf) {
	bool retval;
	normsample_t ac;
	
	retval = samples_max_generic(format, samples, inbuf, &ac);

	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							*((sample_pcm8u_t *) outbuf) = (sample_pcm8u_t) ac;
							break;
						case BIT16:
							*((sample_pcm16u_t *) outbuf) = (sample_pcm16u_t) ac;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							retval = false;
							break;
					}
					break;
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							*((sample_pcm8s_t *) outbuf) = (sample_pcm8s_t) ac;
							break;
						case BIT16:
							*((sample_pcm16s_t *) outbuf) = (sample_pcm16s_t) ac;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							retval = false;
							break;
					}
					break;
				default:
					FATAL("Undefined format.bitencoding");
					retval = false;
					break;
			}
			break;
		default:
			FATAL("Undefined format.encoding");
			retval = false;
			break;
	}
	return retval;
}

normsample_t NormSample(const format_t & format, const void * sample, const normsample_t max) {
	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							return (*((sample_pcm8u_t * ) sample) - sample_pcm8u_0) / (normsample_t) sample_pcm8u_dr * max;
							break;
						case BIT16:
							return (*((sample_pcm16u_t * ) sample) - sample_pcm16u_0) / (normsample_t) sample_pcm16u_dr * max;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							return 0;
					}
					break;
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							return (*((sample_pcm8s_t * ) sample) - sample_pcm8s_0) / (normsample_t) sample_pcm8s_dr * max;
							break;
						case BIT16:
							return (*((sample_pcm16s_t * ) sample) - sample_pcm16s_0) / (normsample_t) sample_pcm16s_dr * max;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							return 0;
					}
					break;
				default:
					FATAL("Undefined format.bitencoding");
					return 0;
			}
			break;
		default:
			FATAL("Undefined format.encoding");
			return 0;
	}
}

bool SamplesAverageNorm(const format_t & format, const samples_t samples, const void * inbuf, normsample_t * average, const normsample_t max) {
	bool retval;
	normsample_t ac;
	
	retval = samples_average_generic(format, samples, inbuf, &ac);
	
	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							*average = (ac - sample_pcm8u_0) / (normsample_t) sample_pcm8u_dr * max;
							break;
						case BIT16:
							*average = (ac - sample_pcm16u_0) / (normsample_t) sample_pcm16u_dr * max;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							retval = false;
							break;
					}
					break;
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							*average = (ac - sample_pcm8s_0) / (normsample_t) sample_pcm8s_dr * max;
							break;
						case BIT16:
							*average = (ac - sample_pcm16s_0) / (normsample_t) sample_pcm16s_dr * max;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							retval = false;
							break;
					}
					break;
				default:
					FATAL("Undefined format.bitencoding");
					retval = false;
					break;
			}
			break;
		default:
			FATAL("Undefined format.encoding");
			retval = false;
			break;
	}
	
	return retval;
}

bool SamplesMaxNorm(const format_t & format, const samples_t samples, const void * inbuf, normsample_t * average, const normsample_t max) {
	bool retval;
	normsample_t ac;
	
	retval = samples_max_generic(format, samples, inbuf, &ac);
	
	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							*average = (ac - sample_pcm8u_0) / (normsample_t) sample_pcm8u_dr * max;
							break;
						case BIT16:
							*average = (ac - sample_pcm16u_0) / (normsample_t) sample_pcm16u_dr * max;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							retval = false;
							break;
					}
					break;
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							*average = (ac - sample_pcm8s_0) / (normsample_t) sample_pcm8s_dr * max;
							break;
						case BIT16:
							*average = (ac - sample_pcm16s_0) / (normsample_t) sample_pcm16s_dr * max;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							retval = false;
							break;
					}
					break;
				default:
					FATAL("Undefined format.bitencoding");
					retval = false;
					break;
			}
			break;
		default:
			FATAL("Undefined format.encoding");
			retval = false;
			break;
	}
	
	return retval;
}

bool ClearSamples(const format_ex_t & format, const samples_t samples, void * buf) {
	samples_t i;

	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							memset(buf, sample_pcm8u_0, sizeof(sample_pcm8u_t) * samples * format.channels);
							break;
						case BIT16:
							for (i = 0; i < samples * format.channels; i++)
								*((sample_pcm16u_t * ) buf + i) = sample_pcm16u_0;
							break;
						default:
							FATAL("Undefined format.samplewidth");
							return false;
					}
					break;
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							memset(buf, sample_pcm8s_0, sizeof(sample_pcm8s_t) * samples * format.channels);
							break;
						case BIT16:
							ASSERT(sample_pcm16s_0 == 0);
							/* To be corrent, we should not use memset() here (the zero amplitude sample
							   could be theoretically not equal to binary zero), but according to the
							   current value of sample_pcm16s_0 this is true.
							   
							   It's just a little optimalization which is affordable. */
							memset(buf, sample_pcm16s_0, sizeof(sample_pcm16s_t) * samples * format.channels);
							break;
						default:
							FATAL("Undefined format.samplewidth");
							return false;
					}
					break;
				default:
					assert("Undefined format.bitencoding");
					return false;
			}
			break;
		default:
			assert("Undefined format.encoding");
			return false;
	}
	
	return true;
}

samples_t SamplesCmp(const format_t & infmt, const format_t & outfmt, const samples_t samples) {
	samples_t mul = 1;
	samples_t div = 1;
	
	if ((infmt.samplewidth == BIT16) && (outfmt.samplewidth == BIT8))
		div *= 2;
		
	if ((infmt.samplewidth == BIT8) && (outfmt.samplewidth == BIT16))
		mul *= 2;

	mul *= infmt.rate;
	div *= outfmt.rate;
	
	return samples * mul / div;		
}

samples_t SamplesCmpIn(const format_t & infmt, const format_t & outfmt, const samples_t insamples) {
	samples_t mul = 1;
	samples_t div = 1;
	
	if ((infmt.samplewidth == BIT16) && (outfmt.samplewidth == BIT8))
		div *= 2;
		
	if ((infmt.samplewidth == BIT8) && (outfmt.samplewidth == BIT16))
		mul *= 2;

	mul *= infmt.rate;
	div *= outfmt.rate;
	
	return insamples * div / mul;		
}

bool CnvSample(const format_t infmt, const format_t outfmt, void * inbuf, void * outbuf) {
	ASSERT(infmt.rate == outfmt.rate);										/* this function doesn't convert sample rates by itself, converts just one sample */
	
	size_t inlen = SampleLen(infmt);
	size_t outlen = SampleLen(outfmt);
	sample_pcm8s_w_t tmp8s;
	sample_pcm8u_w_t tmp8u;
	sample_pcm16s_w_t tmp16s;
	sample_pcm16u_w_t tmp16u;
	
	switch (infmt.samplewidth) {
		case BIT8:
			switch (infmt.bitencoding) {
				case UNSIGNED:
					switch (outfmt.samplewidth) {
						case BIT8:
							switch (infmt.bitencoding) {
								case UNSIGNED:								/* 8b unsigned -> 8b unsigned */
									ASSERT(inlen == outlen);
									memcpy(outbuf, inbuf, inlen);
									break;
								case SIGNED:								/* 8b unsigned -> 8b signed */
									tmp8s = *((sample_pcm8u_t * ) inbuf);
									tmp8s = tmp8s + sample_pcm8u_min - sample_pcm8s_min;
									*((sample_pcm8s_t * ) outbuf) = tmp8s;
									break;
								default:
									FATAL("Undefined outfmt.bitencoding");
									return false;
							}
							break;
						case BIT16:
							switch (infmt.bitencoding) {
								case UNSIGNED:								/* 8b unsigned -> 16b unsigned */
									tmp16u = *((sample_pcm8u_t * ) inbuf);
									tmp16u = ((tmp16u + sample_pcm8u_min) << 8) - sample_pcm16u_min;
									*((sample_pcm16u_t * ) outbuf) = tmp16u;
									break;
								case SIGNED:								/* 8b unsigned -> 16b signed */
									tmp16s = *((sample_pcm8u_t * ) inbuf);
									tmp16s = ((tmp16s + sample_pcm8u_min) << 8) - sample_pcm16s_min;
									*((sample_pcm16s_t * ) outbuf) = tmp16s;
									break;
								default:
									FATAL("Undefined outfmt.bitencoding");
									return false;
							}
							break;
						default:
							FATAL("Undefined outfmt.samplewidth");
							return false;
					}
					break;
				case SIGNED:
					switch (outfmt.samplewidth) {
						case BIT8:
							switch (infmt.bitencoding) {
								case UNSIGNED:								/* 8b signed -> 8b unsigned */
									tmp8u = (*(sample_pcm8u_t * ) inbuf);
									tmp8u = tmp8s + sample_pcm8u_min - sample_pcm8s_min;
									*((sample_pcm8u_t * ) outbuf) = (sample_pcm8u_t) tmp8u;
									break;
								case SIGNED:								/* 8b signed -> 8b signed */
									ASSERT(inlen == outlen);
									memcpy(outbuf, inbuf, inlen);
									break;
								default:
									FATAL("Undefined outfmt.bitencoding");
									return false;
							}
							break;
						case BIT16:
							switch (infmt.bitencoding) {
								case UNSIGNED:								/* 8b signed -> 16b unsigned */
									tmp16u = (*(sample_pcm8s_t * ) inbuf);
									tmp16u = ((tmp16u + sample_pcm8s_min) << 8) - sample_pcm16u_min;
									*((sample_pcm16u_t * ) outbuf) = tmp16u;
									break;
								case SIGNED:								/* 8b signed -> 16b signed */
									tmp16s = (*(sample_pcm8s_t * ) inbuf);
									tmp16s = ((tmp16s + sample_pcm8s_min) << 8) - sample_pcm16s_min;
									*((sample_pcm16s_t * ) outbuf) = tmp16s;
									break;
								default:
									FATAL("Undefined outfmt.bitencoding");
									return false;
							}
							break;
						default:
							FATAL("Undefined outfmt.samplewidth");
							return false;
					}
					break;
				default:
					assert("Undefined infmt.bitencoding");
					break;
			}
			break;
		case BIT16:
			switch (infmt.bitencoding) {
				case UNSIGNED:
					switch (outfmt.samplewidth) {
						case BIT8:
							switch (infmt.bitencoding) {
								case UNSIGNED:								/* 16b unsigned -> 8b unsigned */
									tmp16u = (*(sample_pcm16u_t * ) inbuf);
									tmp8u = ((tmp16u + sample_pcm16u_min) >> 8) - sample_pcm8u_min;
									*((sample_pcm8u_t * ) outbuf) = tmp8u;
									break;
								case SIGNED:								/* 16b unsigned -> 8b signed */
									tmp16u = (*(sample_pcm16u_t * ) inbuf);
									tmp8s = ((tmp16u + sample_pcm16u_min) >> 8) - sample_pcm8s_min;
									*((sample_pcm8s_t * ) outbuf) = tmp8s;
									break;
								default:
									FATAL("Undefined outfmt.bitencoding");
									return false;
							}
							break;
						case BIT16:
							switch (infmt.bitencoding) {
								case UNSIGNED:								/* 16b unsigned -> 16b unsigned */
									ASSERT(inlen == outlen);
									memcpy(outbuf, inbuf, inlen);
									break;
								case SIGNED:								/* 16b unsigned -> 16b signed */
									tmp16s = (*(sample_pcm16u_t * ) inbuf);
									tmp16s = ((tmp16s + sample_pcm16u_min) >> 8) - sample_pcm16s_min;
									*((sample_pcm16s_t * ) outbuf) = tmp16s;
									break;
								default:
									FATAL("Undefined outfmt.bitencoding");
									return false;
							}
							break;
						default:
							FATAL("Undefined outfmt.samplewidth");
							return false;
					}
					break;
				case SIGNED:
					switch (outfmt.samplewidth) {
						case BIT8:
							switch (infmt.bitencoding) {
								case UNSIGNED:								/* 16b signed -> 8b unsigned */
									tmp16s = (*(sample_pcm16s_t * ) inbuf);
									tmp8u = ((tmp16s + sample_pcm16s_min) >> 8) - sample_pcm8u_min;
									*((sample_pcm8u_t * ) outbuf) = tmp8u;
									break;
								case SIGNED:								/* 16b signed -> 8b signed */
									tmp16s = (*(sample_pcm16s_t * ) inbuf);
									tmp8s = ((tmp16s + sample_pcm16s_min) >> 8) - sample_pcm8s_min;
									*((sample_pcm8s_t * ) outbuf) = tmp8s;
									break;
								default:
									FATAL("Undefined outfmt.bitencoding");
									return false;
							}
							break;
						case BIT16:
							switch (infmt.bitencoding) {
								case UNSIGNED:								/* 16b signed -> 16b unsigned */
									tmp16s = (*(sample_pcm16s_t * ) inbuf);
									tmp16u = ((tmp16s + sample_pcm16s_min) >> 8) - sample_pcm16u_min;
									*((sample_pcm16u_t * ) outbuf) = tmp16u;
									break;
								case SIGNED:								/* 16b signed -> 16b signed */
									ASSERT(inlen == outlen);
									memcpy(outbuf, inbuf, inlen);
									break;
								default:
									FATAL("Undefined outfmt.bitencoding");
									return false;
							}
							break;
						default:
							FATAL("Undefined outfmt.samplewidth");
							return false;
					}
					break;
				default:
					assert("Undefined infmt.bitencoding");
					break;
			}
			break;
		default:
			FATAL("Undefined infmt.samplewidth");
			return false;
	}
	
	return true;
}

samples_t SamplesConv(const format_t & infmt, const format_ex_t & outfmt, const samples_t samples, void * inbuf, void * outbuf) {
	format_t ifmt = infmt;								/* input format in progress */
	void * ibuf = inbuf;								/* progress buffer */
	bool icpy = true;									/* is ibuf pointing to the real input buffer */
	samples_t isamples = samples;						/* number of samples in ibuf */
	samples_t i;
	
	if ((outfmt.samplewidth == BIT16) && (infmt.samplewidth == BIT8)) {					/* do the conversion to higher resolution first */
		ifmt.encoding = infmt.encoding;
		ifmt.bitencoding = outfmt.bitencoding;
		ifmt.samplewidth = outfmt.samplewidth;
		ifmt.rate = infmt.rate;
		ibuf = AllocSamples(ifmt, samples);
		// FIXME chech ibuf for NULL
		icpy = false;
		for (i = 0; i < samples; i++) {
			void * orig = GetSample(infmt, i, inbuf);
			void * cnv = GetSample(ifmt, i, ibuf);
			CnvSample(infmt, ifmt, orig, cnv);	// FIXME check for return value
		}
	}
	
	// FIXME: do sample rate conversion here
	
	// FIXME: do final encoding conversion here
	
	if (outfmt.channels > 1) {							/* replicate samples */
		for (i = 0; i < isamples; i++) {
			void * tmp = GetSample(ifmt, i, ibuf);
			for (index_t j = 0; j < outfmt.channels; j++) {
				void * dup = GetSample(outfmt, j, i, outbuf);
				memcpy(dup, tmp, SampleLen(ifmt));				
			}
		}
	} else												/* no channel replication, just copy */
		memcpy(outbuf, ibuf, isamples * SampleLen(ifmt));
		
	if (!icpy)
		FreeSamples(ifmt, ibuf);
		
	return isamples;
}

bool AddSamples(const format_ex_t & format, const samples_t tsamples, const samples_t samples, const void * tbuf, const void * buf) {
	bool clip = false;
	samples_t i;

	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							for (i = 0; i < tsamples * format.channels; i++) {
								sample_pcm8u_t orig = *((sample_pcm8u_t * ) buf + i);
								sample_pcm8u_t onew = *((sample_pcm8u_t * ) tbuf + i);
								
								sample_pcm8u_w_t res = orig + onew;
								
								if (res > sample_pcm8u_max) {
									res = sample_pcm8u_max;
									clip = true;
								}
								if (res < sample_pcm8u_min) {
									res = sample_pcm8u_min;
									clip = true;
								}
								
								*((sample_pcm8u_t * ) buf + i) = res;
							}
							break;
						case BIT16:
							for (i = 0; i < tsamples * format.channels; i++) {
								sample_pcm16u_t orig = *((sample_pcm16u_t * ) buf + i);
								sample_pcm16u_t onew = *((sample_pcm16u_t * ) tbuf + i);
								
								sample_pcm16u_w_t res = orig + onew;
								
								if (res > sample_pcm16u_max) {
									res = sample_pcm16u_max;
									clip = true;
								}
								if (res < sample_pcm16u_min) {
									res = sample_pcm16u_min;
									clip = true;
								}
								
								*((sample_pcm16u_t * ) buf + i) = res;
							}
							break;
						default:
							FATAL("Undefined format.samplewidth");
							break;
					}
					break;
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							for (i = 0; i < tsamples * format.channels; i++) {
								sample_pcm8s_t orig = *((sample_pcm8s_t * ) buf + i);
								sample_pcm8s_t onew = *((sample_pcm8s_t * ) tbuf + i);
								
								sample_pcm8s_w_t res = orig + onew;
								
								if (res > sample_pcm8s_max) {
									res = sample_pcm8s_max;
									clip = true;
								}
								if (res < sample_pcm8s_min) {
									res = sample_pcm8s_min;
									clip = true;
								}
								
								*((sample_pcm8s_t * ) buf + i) = res;
							}
							break;
						case BIT16:
							for (i = 0; i < tsamples * format.channels; i++) {
								sample_pcm16s_t orig = *((sample_pcm16s_t * ) buf + i);
								sample_pcm16s_t onew = *((sample_pcm16s_t * ) tbuf + i);
								
								sample_pcm16s_w_t res = orig + onew;
								
								if (res > sample_pcm16s_max) {
									res = sample_pcm16s_max;
									clip = true;
								}
								if (res < sample_pcm16s_min) {
									res = sample_pcm16s_min;
									clip = true;
								}
								
								*((sample_pcm16s_t * ) buf + i) = res;
							}
							break;
						default:
							FATAL("Undefined format.samplewidth");
							break;
					}
					break;
				default:
					FATAL("Undefined format.bitencoding");
					break;
			}
			break;
		default:
			FATAL("Undefined format.encoding");
			break;
	}
	
	return clip;
}

void FormatAlias(const format_ex_t & infmt, format_t & outfmt) {
	outfmt.encoding = infmt.encoding;
	outfmt.bitencoding = infmt.bitencoding;
	outfmt.samplewidth = infmt.samplewidth;
	outfmt.rate = infmt.rate;
}

