/***********************************************************************

  FinalMix: generic.h
  by Martin Decky
  
  Redistribution of this file is permitted under the GNU Public License.

  0.9.0		Initial version
  0.9.1		Code cleanup
  0.9.2		Standard WX types
  0.9.3		Standardized types
  0.9.4     Cleanup

***********************************************************************/

#ifndef _FM_GENERIC_H_
#define _FM_GENERIC_H_

#define DEBUG 1

#include <wx/wx.h>

#define FM_NAME "FinalMix"
#define FM_NAME_UPPER "FINALMIX"
#define FM_VERSION_STRING "0.9.4"

#define FM_PEAK_TAG "FMPE"
#define FM_PROJECT_TAG "FMPR"

#define FM_PEAK_EXT "fpe"

#define BUF_SIZE 204800

#define DEF_PEAK_DIR "."
#define DEF_PEAK_RATIO 1000

#define FULL_PERCENT 100

/** Function return status type.
 *
 */

enum status_t {
	OK = 0,									/**< No error */
	IO_ERROR = -1,							/**< Input/output error */
	UNK_ERROR = -2,							/**< Unknown error */
	UNK_FORMAT = -3,						/**< Unknown file format */
	STRUCT_ERR = -4,						/**< Error in file structure */
	UNS_FORMAT = -5							/**< Unsupported file format */
};

/** Sample encoding type.
 *
 */

enum encoding_t {
	PCM = 1									/**< PCM */
};

/** Bitwise encoding of sample type.
 *
 */

enum bitencoding_t {
	UNSIGNED = 1,							/**< Unsigned */
	SIGNED									/**< Signed */
};

/** Bit width of sample type.
 *
 */

enum samplewidth_t {
	BIT16 = 1,								/**< 16 bits */
	BIT8									/**< 8 bits */
};

typedef wxUint8 sample_pcm8u_t;				/**< Sample type PCM 8 bit unsigned */
typedef wxInt8 sample_pcm8s_t;				/**< Sample type PCM 8 bit signed */
typedef wxUint16 sample_pcm16u_t;			/**< Sample type PCM 16 bit unsigned */
typedef wxInt16 sample_pcm16s_t;			/**< Sample type PCM 16 bit signed */

typedef wxUint16 sample_pcm8u_w_t;			/**< Sample type PCM 8 bit unsigned (double precision) */
typedef wxInt16 sample_pcm8s_w_t;			/**< Sample type PCM 8 bit signed (double precision) */
typedef wxUint32 sample_pcm16u_w_t;			/**< Sample type PCM 16 bit unsigned (double precision) */
typedef wxInt32 sample_pcm16s_w_t;			/**< Sample type PCM 16 bit signed (double precision) */

typedef wxUint32 samplerate_t;				/**< Typ pro uchovn vzorkovac frekvence */
typedef wxUint64 samples_t;					/**< Typ pro uchovn potu vzork */
typedef wxInt64 samples_ex_t;				/**< Typ pro mezikonverzi poctu vzorku na double */

/** Sample precision constants.
 *
 */

extern const sample_pcm8u_t sample_pcm8u_min;
extern const sample_pcm8u_t sample_pcm8u_max;
extern const sample_pcm8u_t sample_pcm8u_0;

extern const sample_pcm8s_t sample_pcm8s_min;
extern const sample_pcm8s_t sample_pcm8s_max;
extern const sample_pcm8s_t sample_pcm8s_0;

extern const sample_pcm16u_t sample_pcm16u_min;
extern const sample_pcm16u_t sample_pcm16u_max;
extern const sample_pcm16u_t sample_pcm16u_0;

extern const sample_pcm16s_t sample_pcm16s_min;
extern const sample_pcm16s_t sample_pcm16s_max;
extern const sample_pcm16s_t sample_pcm16s_0;

typedef wxUint16 version_t;

extern const version_t fmfile_version_major;	/**< Major version number in project file header */
extern const version_t fmfile_version_minor;	/**< Minor version number in project file header */

/** One mono track format structure.
 *
 */

typedef struct {
	enum encoding_t encoding;				/**< Encoding */
	enum bitencoding_t bitencoding;			/**< Bit encoding */
	enum samplewidth_t samplewidth;			/**< Sample width */
	samplerate_t rate;						/**< Sample rate */
	samples_t samples;						/**< Number of samples */
} format_t;

typedef int index_t;						/**< Internal (platform-specific) indexation type */
typedef wxUint64 length_t;					/**< Maximal length type */
typedef wxFloat64 normsample_t;

/** Input/output channel format structure.
 *
 */

typedef struct {
	enum encoding_t encoding;				/**< Encoding */
	enum bitencoding_t bitencoding;			/**< Bit encoding */
	enum samplewidth_t samplewidth;			/**< Sample width */
	samplerate_t rate;						/**< Sample rate */
	index_t channels;						/**< Number of interleaved channels */
} format_ex_t;

typedef wxUint16 track_t;					/**< Channel or track indexing type */
typedef wxUint32 ratio_t;					/**< Display ratio type */

/** Global Fatal Error.
 *
 * This procedure is used to indicate fatal or unreachable situations in code.
 * Displays error message on standard error output and halts the application
 * if in debug mode (i.e. DEBUG macro defined).
 *
 * @param format Standard format string used by fprintf().
 *
 */

void Fatal(const char * format, ...);

/** Global Assert.
 *
 * Checks the value of the expression and if it is false, raises a fatal
 * error using Fatal.
 *
 * The test is active.
 *
 * @param expr Condition which should be true.
 *
 */

#if defined(__WXMSW__)
#define __FUNCTION__ "unknown"
#endif

#define ASSERT(expr) if (!(expr)) { Fatal("Assertion failed: %s (in %s() at %s on line %d)", #expr, __FUNCTION__, __FILE__, __LINE__); }

/** Global Unconditional Assert.
 *
 * Raises Fatal unconditionaly, but displays the same error description
 * as ASSERT.
 *
 * @param expr Description of the error.
 *
 */

#define FATAL(desc) { Fatal("Assertion failed: %s (in %s() at %s on line %d)", desc, __FUNCTION__, __FILE__, __LINE__); }

/** Construct filename from components.
 *
 * Constructs filename with path (in standard wxWindows representation).
 * Extension and path is chopped from the name of the file (second parameter),
 * the new path is prepended at the beginning and the new suffix and extension
 * is appended.
 *
 * @param path    New filename path (absolute or relative).
 * @param name    Original filename (path and extension is chopped).
 * @param ext     New filename extension.
 * @param fsuffix Filename suffix (placed between the bare name and new extension).
 *
 * @return Constructed filename.
 *
 */

wxString SetFilename(const wxString path, const wxString name, const wxString ext, const wxString fsuffix);

/** Size of one sample.
 *
 * Returns the size in bytes of one sample in given format. Raises Fatal on unknown format.
 *
 * @param format Sample format description.
 *
 * @return Size of one sample in bytes, in case of format_ex_t the returned size is not referencing
 *         to the number of channels.
 *
 */

size_t SampleLen(const format_t & format);
size_t SampleLen(const format_ex_t & format);

/** Allocates memory for the given number of samples.
 *
 * Dynamicaly allocates memory for storing given number of samples in given format.
 * Raises Fatal on unknown format.
 *
 * @param format Sample format description, in case of format_ex_t the number of channels
 *               in the structure is ignored.
 * @param count  Number of samples to allocate.
 *
 * @return Pointer to the new allocated memory, NULL on fail.
 *
 */

void * AllocSamples(const format_t & format, const samples_t count);
void * AllocSamples(const format_ex_t & format, const samples_t count);

/** Safely deallocates memory for samples.
 *
 * Safely deallocates memory previously allocated by AllocSamples.
 * Raises Fatal on unknown format, other error states indicates
 * in the same way as delete operator.
 *
 * @param format Sample format description, in case of format_ex_t the number of channels
 *               in the structure is ignored.
 * @param data   Pointer to the memory allocated by AllocSamples.
 *
 * @return Return true on sucess.
 *
 */

bool FreeSamples(const format_t & format, const void * data);

/** Returns average sample.
 *
 * Calculates the average value of all samples from input buffer. Raises Fatal on unknown format.
 *
 * @param format  Sample format description.
 * @param samples Number of samples in inbuf.
 * @param inbuf   Input buffer.
 * @param outbuf  Output buffer, which must be preallocated to hold at least one sample
 *                in the format described by the first parameter.
 *
 * @return Returns true on succesful calculation.
 *
 */

bool SamplesAverage(const format_t & format, const samples_t samples, const void * inbuf, const void * outbuf);
bool SamplesMax(const format_t & format, const samples_t samples, const void * inbuf, const void * outbuf);

/** Returns normalized value of the sample (in internal format).
 *
 * Calculates the value of the sample in internal representation (zero amplitude is 0, maximal amplitude
 * is +max, maximal negative amplitude is -max). Raises Fatal on unknown format.
 *
 * @param format Sample format description.
 * @param sample Input sample.
 * @param max    Maximal value of linear PCM amplitude. If the value is negative, the resulting
 *               amplitude is phase reversed.
 *
 * @return Normalized PCM amplitude in the range from -max to +max.
 *
 */

normsample_t NormSample(const format_t & format, const void * sample, const normsample_t max);

/** Returns pointer to one single sample in buffer.
 *
 * Returns the pointer of a sample indexed by pos in buffer. Raises Fatal on unknown format.
 *
 * @param format  Sample format description.
 * @param pos     Position of the sample in the buffer (in samples, zero based).
 * @param samples Input buffer.
 *
 * @return Pointer to the sample (in the original buffer, not a copy).
 *
 */
 
/** Returns average sample in normalized internal format.
 *
 * Calculates the average value of all samples from input buffer. Raises Fatal on unknown format.
 *
 * @param format  Sample format description.
 * @param samples Number of samples in inbuf.
 * @param inbuf   Input buffer.
 * @param average Average sample in normalized format.
 * @param max     Maximal value of linear PCM amplitude. If the value is negative, the resulting
 *                amplitude is phase reversed.
 *
 * @return Returns true on succesful calculation.
 *
 */

bool SamplesAverageNorm(const format_t & format, const samples_t samples, const void * inbuf, double * average, const normsample_t max);
bool SamplesMaxNorm(const format_t & format, const samples_t samples, const void * inbuf, double * average, const normsample_t max);

/** Returns normalized value of the sample (in internal format).
 *
 * Calculates the value of the sample in internal representation (zero amplitude is 0, maximal amplitude
 * is +max, maximal negative amplitude is -max). Raises Fatal on unknown format.
 *
 * @param format Sample format description.
 * @param sample Input sample.
 * @param max    Maximal value of linear PCM amplitude. If the value is negative, the resulting
 *               amplitude is phase reversed.
 *
 * @return Normalized PCM amplitude in the range from -max to +max.
 *
 */

inline void * GetSample(const format_t & format, const samples_t pos, const void * samples) {
	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							return (sample_pcm8u_t * ) samples + pos;
						case BIT16:
							return (sample_pcm16u_t * ) samples + pos;
						default:
							FATAL("Undefined format.samplewidth");
							return NULL;
					}
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							return (sample_pcm8s_t * ) samples + pos;
						case BIT16:
							return (sample_pcm16s_t * ) samples + pos;
						default:
							FATAL("Undefined format.samplewidth");
							return NULL;
					}
				default:
					FATAL("Undefined format.bitencoding");
					return NULL;
			}
		default:
			FATAL("Undefined format.encoding");
			return NULL;
	}
}

inline void * GetSample(const format_ex_t & format, const index_t chan, const samples_t pos, const void * samples) {
	switch (format.encoding) {
		case PCM:
			switch (format.bitencoding) {
				case UNSIGNED:
					switch (format.samplewidth) {
						case BIT8:
							return (sample_pcm8u_t * ) samples + pos * format.channels + chan;
						case BIT16:
							return (sample_pcm16u_t * ) samples + pos * format.channels + chan;
						default:
							FATAL("Undefined format.samplewidth");
							return NULL;
					}
				case SIGNED:
					switch (format.samplewidth) {
						case BIT8:
							return (sample_pcm8s_t * ) samples + pos * format.channels + chan;
						case BIT16:
							return (sample_pcm16s_t * ) samples + pos * format.channels + chan;
						default:
							FATAL("Undefined format.samplewidth");
							return NULL;
					}
				default:
					FATAL("Undefined format.bitencoding");
					return NULL;
			}
		default:
			FATAL("Undefined format.encoding");
			return NULL;
	}
}

/** Converts number of samples from one format to another.
 *
 * Converts the number of samples in one sample format to the number of samples that would
 * be needed after conversion to another sample format. The calculation is based on
 * encoding, bit encoding, sample width and sample rate. The returned value is consistent
 * with the output buffer requirements of SampleConv (both local and global conversion).
 * Raises Fatal on unknown format.
 *
 * @param infmt   Input format description.
 * @param outfmt  Output format description.
 * @param samples Number of samples in input format.
 *
 * @return Number of samples in output format.
 *
 */

samples_t SamplesCmp(const format_t & infmt, const format_t & outfmt, const samples_t samples);
samples_t SamplesCmpIn(const format_t & infmt, const format_t & outfmt, const samples_t insamples);

/** Local coversion between various formats.
 *
 * Converts audio data between given formats. The conversion is done using the
 * nearest neibourghood algorithm w/o antialiasing, which makes it suitable
 * for both global and local (partial) conversion. The conversion is based on
 * encoding, bit encoding, sample width and sample rate. The function can
 * also duplicate the output data to multiple output channels if needed.
 * Sample rate is converted always in the more precise format of both input
 * and output. Channel multiplication is done always as the final step.
 * Raises Fatal on unknown format.
 *
 * @param infmt   Input format description.
 * @param outfmt  Output format description.
 * @param samples Number of input samples.
 * @param inbuf   Input buffer.
 * @param outbuf  Output buffer, must be preallocated to hold at least the size of samples
 *                returned by SamplesCmp.
 *
 * @return Returns the number of output samples.
 *
 */

samples_t SamplesConv(const format_t & infmt, const format_ex_t & outfmt, const samples_t samples, void * inbuf, void * outbuf);

/** Zero modulation.
 *
 * Sets the buffer to zero amplitude. Raises Fatal on unknown format.
 *
 * @param format  Format description.
 * @param samples Number of samples to zero.
 * @param buf     Output buffer, must be preallocated to hold at least all zero samples.
 *
 * @return Return true on sucess.
 *
 */

bool ClearSamples(const format_ex_t & format, const samples_t samples, void * buf);

/** Add additional samples to given samples.
 *
 * This function adds additional samples to the samples already present in the buffer,
 * practically performing an overlay-mix. The samples are simply added together, however
 * underflow or overflow is checked and handled by cropping (setting maximal or minimal valid value).
 * The return value indicates whether the overflow or underflow occured.
 * Raises Fatal on unknown format.
 *
 * @param format   Format description.
 * @param tsamples Number of samples of the new audio data.
 * @param samples  Number of samples in the input/output buffer.
 * @param tbuf     New audio data.
 * @param buf      Input/output buffer.
 *
 * @return Returns true if overflow or underflow occured.
 *
 */
 
void FormatAlias(const format_ex_t & infmt, format_t & outfmt);

bool AddSamples(const format_ex_t & format, const samples_t tsamples, const samples_t samples, const void * tbuf, const void * buf);

/** Minimum of two values.
 *
 * Returns the smallest value (according to operator <) of two inputs of the same type. The operator
 * < must be defined for this type.
 *
 * @param a First value.
 * @param b Second value.
 *
 * @return Minimum of the two values.
 */

template< class T > inline T min(const T a, const T b) {
	if (a < b)
		return a;
	else
		return b;
}

template< class T > inline T max(const T a, const T b) {
	if (a > b)
		return a;
	else
		return b;
}

#endif
